#include <Drivers.h>
#include <KernelExport.h>
#include <CAM.h>
#include <string.h>
#include "ct.h"

/*
#define DEBUG 1
*/

device_entry devices[8]=
{
	{ "/dev/scsi_tape_000",ct_open,ct_close,ct_control,ct_read,ct_write },
	{ "/dev/scsi_tape_010",ct_open,ct_close,ct_control,ct_read,ct_write },
	{ "/dev/scsi_tape_020",ct_open,ct_close,ct_control,ct_read,ct_write },
	{ "/dev/scsi_tape_030",ct_open,ct_close,ct_control,ct_read,ct_write },
	{ "/dev/scsi_tape_040",ct_open,ct_close,ct_control,ct_read,ct_write },
	{ "/dev/scsi_tape_050",ct_open,ct_close,ct_control,ct_read,ct_write },
	{ "/dev/scsi_tape_060",ct_open,ct_close,ct_control,ct_read,ct_write },
	0
};

long cam_action_sync (CCB_HEADER *ccb)
{
	long status=xpt_action (ccb);
	
	if (status==CAM_SUCCESS)
	{
		while (ccb->cam_status==CAM_REQ_INPROG)
		{
		}		
	}
	
	return status;
}

long cam_fmt_ccb_header (CCB_HEADER *ccb, u_char func_code, u_char path_id, u_char target_id, u_char target_lun, U32 flags)
{
	u_short ccb_length;
		
	switch (func_code)
	{
		case XPT_NOOP:
			return B_ERROR;
		case XPT_SCSI_IO:
			ccb_length=sizeof(CCB_SCSIIO);
			break;
		case XPT_GDEV_TYPE:
			ccb_length=sizeof(CCB_GETDEV);
			break;
		case XPT_PATH_INQ:
			ccb_length=sizeof(CCB_PATHINQ);
			break;		
		case XPT_REL_SIMQ:
			ccb_length=sizeof(CCB_RELSIM);
			break;
		case XPT_SASYNC_CB:
			ccb_length=sizeof(CCB_SETASYNC);
			break;
		case XPT_SDEV_TYPE:		
			ccb_length=sizeof(CCB_SETDEV);
			break;
		case XPT_SCAN_BUS:
			return B_ERROR;
		case XPT_ABORT:
			ccb_length=sizeof(CCB_ABORT);
			break;
		case XPT_RESET_BUS:
			ccb_length=sizeof(CCB_RESETBUS);
			break;
		case XPT_RESET_DEV:
			ccb_length=sizeof(CCB_RESETDEV);
			break;
		case XPT_TERM_IO:
			ccb_length=sizeof(CCB_TERMIO);
			break;
		case XPT_ENG_INQ:
			ccb_length=sizeof(CCB_ENG_INQ);
			break;
		case XPT_ENG_EXEC:
			ccb_length=sizeof(CCB_ENG_EXEC);
			break;
		case XPT_EN_LUN:
			ccb_length=sizeof(CCB_ENABLE_LUN);
			break;
		case XPT_TARGET_IO:
			return B_ERROR;
		case XPT_ACCEPT_TARG:
			ccb_length=sizeof(CCB_ACCEPT_TARG);
			break;
		case XPT_CONT_TARG:
			return B_ERROR;
		case XPT_IMMED_NOTIFY:
			ccb_length=sizeof(CCB_IMMED_NOTIFY);
			break;
		case XPT_NOTIFY_ACK:
			ccb_length=sizeof(CCB_NOTIFY_ACK);
			break;
		case XPT_FUNC:
			return B_ERROR;
		case XPT_VUNIQUE:
			return B_ERROR;
			
		default:
			return B_ERROR;
	}

	ccb->my_addr=ccb;
	ccb->cam_ccb_len=ccb_length;	
	ccb->cam_func_code=func_code;	
	ccb->cam_status=CAM_REQ_INPROG;
	ccb->cam_path_id=path_id;
	ccb->cam_target_id=target_id;
	ccb->cam_target_lun=target_lun;
	ccb->cam_flags=flags;

	return B_NO_ERROR;
}

long cam_do_scsi_noio (u_char path_id, u_char target_id, u_char target_lun, struct cdb6 *cdb, u_char *cam_status, u_char *scsi_status)
{
	long status=B_ERROR;
	struct sense_data sensebuffer;
	CCB_SCSIIO *ccb=(CCB_SCSIIO *)xpt_ccb_alloc();
	
	if (ccb!=NULL)
	{	
		status=cam_fmt_ccb_header ((CCB_HEADER *)ccb,XPT_SCSI_IO,path_id,target_id,target_lun,CAM_DIR_NONE | CAM_DIS_CALLBACK | CAM_CDB_POINTER);
		
		if (status==B_NO_ERROR)
		{
			ccb->cam_sense_ptr=(u_char *)&sensebuffer;
			ccb->cam_sense_len=sizeof(sensebuffer);

			ccb->cam_cdb_len=6;
			ccb->cam_cdb_io.cam_cdb_ptr=(u_char *)cdb;			
			
			status=cam_action_sync ((CCB_HEADER *)ccb);
			
			*cam_status=ccb->cam_ch.cam_status;
			*scsi_status=((CCB_SCSIIO *)ccb)->cam_scsi_status;
				
			if (status==CAM_SUCCESS)
				status=B_NO_ERROR;
			else
				status=B_ERROR;
		}
	
		xpt_ccb_free(ccb);
	}
	
	if ((*cam_status) & 0x80)
		outputautosensebuffer(&sensebuffer);
	
	*cam_status&=0x7f;

	return status;
}

long cam_do_scsi_output (u_char path_id, u_char target_id, u_char target_lun, struct cdb6 *cdb, u_char *buffer, u_long buflen, u_char *cam_status, u_char *scsi_status)
{
	long status=B_ERROR;
	struct sense_data sensebuffer;
	CCB_SCSIIO *ccb;

	status=lock_memory (buffer,buflen,FALSE);
	
	if (status!=B_NO_ERROR)
	{
		ddprint ("could not lock memory.\n");
		return B_ERROR;
	}	
	
	ccb=(CCB_SCSIIO *)xpt_ccb_alloc();
		
	if (ccb!=NULL)
	{	
		status=cam_fmt_ccb_header ((CCB_HEADER *)ccb,XPT_SCSI_IO,path_id,target_id,target_lun,CAM_DIR_OUT | CAM_DIS_CALLBACK | CAM_CDB_POINTER);
		
		if (status==B_NO_ERROR)
		{
			ccb->cam_data_ptr=buffer;
			ccb->cam_dxfer_len=buflen;
			
			ccb->cam_sense_ptr=(u_char *)&sensebuffer;
			ccb->cam_sense_len=sizeof(sensebuffer);
			
			ccb->cam_cdb_len=6;
			ccb->cam_cdb_io.cam_cdb_ptr=(u_char *)cdb;			
			
			status=cam_action_sync ((CCB_HEADER *)ccb);
			
			*cam_status=ccb->cam_ch.cam_status;
			*scsi_status=((CCB_SCSIIO *)ccb)->cam_scsi_status;

			if (status==CAM_SUCCESS)
				status=B_NO_ERROR;
			else
				status=B_ERROR;
		}
	
		xpt_ccb_free(ccb);
	}

	if ((*cam_status) & 0x80)
		outputautosensebuffer(&sensebuffer);
	
	*cam_status&=0x7f;

	status=unlock_memory (buffer,buflen);

	if (status!=B_NO_ERROR)
		ddprint ("could not unlock memory.\n");
					
	return status;
}

long cam_do_scsi_input (u_char path_id, u_char target_id, u_char target_lun, struct cdb6 *cdb, u_char *buffer, u_long buflen, u_char *cam_status, u_char *scsi_status)
{
	long status=B_ERROR;
	struct sense_data sensebuffer;
	CCB_SCSIIO *ccb;

	status=lock_memory (buffer,buflen,TRUE);
	
	if (status!=B_NO_ERROR)
	{
		ddprint ("could not lock memory.\n");
		return B_ERROR;
	}	
	
	ccb=(CCB_SCSIIO *)xpt_ccb_alloc();
		
	if (ccb!=NULL)
	{	
		status=cam_fmt_ccb_header ((CCB_HEADER *)ccb,XPT_SCSI_IO,path_id,target_id,target_lun,CAM_DIR_IN | CAM_DIS_CALLBACK | CAM_CDB_POINTER);
		
		if (status==B_NO_ERROR)
		{
			ccb->cam_data_ptr=buffer;
			ccb->cam_dxfer_len=buflen;
			
			ccb->cam_sense_ptr=(u_char *)&sensebuffer;
			ccb->cam_sense_len=sizeof(sensebuffer);
			
			ccb->cam_cdb_len=6;
			ccb->cam_cdb_io.cam_cdb_ptr=(u_char *)cdb;			
			
			status=cam_action_sync ((CCB_HEADER *)ccb);
			
			*cam_status=ccb->cam_ch.cam_status;
			*scsi_status=((CCB_SCSIIO *)ccb)->cam_scsi_status;

			if (status==CAM_SUCCESS)
				status=B_NO_ERROR;
			else
				status=B_ERROR;
		}
	
		xpt_ccb_free(ccb);
	}

	if ((*cam_status) & 0x80)
		outputautosensebuffer(&sensebuffer);
	
	*cam_status&=0x7f;

	status=unlock_memory (buffer,buflen);

	if (status!=B_NO_ERROR)
		ddprint ("could not unlock memory.\n");
					
	return status;
}

long cam_readblocklimits (u_char path_id, u_char target_id, u_char target_lun,u_long *min , u_long *max, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;

	struct
	{
		u_char reserved;
		u_char max0;
		u_char max1;
		u_char max2;
		u_char min0;
		u_char min1;
	} limitdata;

	long status;
		
	cdb.opcode=SCSI_READ_BLOCK_LIMITS;
	cdb.lun=(target_lun<<5);
	cdb.byte3=0;
	cdb.byte4=0;
	cdb.length=0;		
	cdb.control=0;

	status=cam_do_scsi_input (path_id,target_id,target_lun,&cdb,(u_char *)&limitdata,sizeof(limitdata),cam_status,scsi_status);	

	*min=(limitdata.min0<<8)|(limitdata.min1);
	*max=(limitdata.max0<<16)|(limitdata.max1<<8)|(limitdata.max2);

	return status;
}

long cam_read (u_char path_id, u_char target_id, u_char target_lun, u_char *buffer, u_long buflen, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;
	
	cdb.opcode=SCSI_READ;
	cdb.lun=(target_lun<<5);
	cdb.byte3=((buflen & 0xff0000)>>16);
	cdb.byte4=((buflen & 0xff00)>>8);
	cdb.length=(buflen & 0xff);		
	cdb.control=0;

	return cam_do_scsi_input (path_id,target_id,target_lun,&cdb,buffer,buflen,cam_status,scsi_status);
}

long cam_write (u_char path_id, u_char target_id, u_char target_lun, u_char *buffer, u_long buflen, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;
	
	cdb.opcode=SCSI_WRITE;
	cdb.lun=(target_lun<<5);
	cdb.byte3=((buflen & 0xff0000)>>16);
	cdb.byte4=((buflen & 0xff00)>>8);
	cdb.length=(buflen & 0xff);		
	cdb.control=0;

	return cam_do_scsi_output (path_id,target_id,target_lun,&cdb,buffer,buflen,cam_status,scsi_status);
}

long cam_tur (u_char path_id, u_char target_id, u_char target_lun, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;
	
	cdb.opcode=SCSI_TUR;
	cdb.lun=target_lun;
	cdb.byte3=cdb.byte4=cdb.length=0;
	cdb.control=0;

	return cam_do_scsi_noio (path_id,target_id,target_lun,&cdb,cam_status,scsi_status);
}
		
long cam_rewind (u_char path_id, u_char target_id, u_char target_lun, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;
	
	cdb.opcode=SCSI_REWIND;
	cdb.lun=(target_lun<<5) | 1;
	cdb.byte3=cdb.byte4=cdb.length=0;
	cdb.control=0;

	return cam_do_scsi_noio (path_id,target_id,target_lun,&cdb,cam_status,scsi_status);
}

long cam_loadunload (u_char path_id, u_char target_id, u_char target_lun, bool load, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;
	
	cdb.opcode=SCSI_LOADUNLOAD;
	cdb.lun=(target_lun<<5) | 1;
	cdb.byte3=0;
	cdb.byte4=0;
	cdb.length=load;
	cdb.control=0;

	return cam_do_scsi_noio (path_id,target_id,target_lun,&cdb,cam_status,scsi_status);
}

long cam_erase (u_char path_id, u_char target_id, u_char target_lun, bool full, u_char *cam_status, u_char *scsi_status)
{
	struct cdb6 cdb;
	
	cdb.opcode=SCSI_ERASE;
	cdb.lun=(target_lun<<5) | full | 2;
	cdb.byte3=0;
	cdb.byte4=0;
	cdb.length=0;
	cdb.control=0;

	return cam_do_scsi_noio (path_id,target_id,target_lun,&cdb,cam_status,scsi_status);
}

long ct_open (device_info *info, ulong flags)
{
	u_char path_id,target_id,target_lun;
	u_char cam_status,scsi_status;
	ulong minblksize,maxblksize;	
	long status;
	
	path_id=(info->entry->name)[15]-'0';
	target_id=(info->entry->name)[16]-'0';
	target_lun=(info->entry->name)[17]-'0';
		
	ddclear();

	ddprint ("open\n");
		
	ddprint ("test unit ready...");
		
	info->private_data=NULL;
		
	status=cam_tur (path_id,target_id,target_lun,&cam_status,&scsi_status);

	if ((status!=B_NO_ERROR)||(cam_status!=CAM_REQ_CMP)||(scsi_status!=0))
		return B_ERROR;

	ddprint ("successful\n");
	ddprint ("read blocklimits...");
							
	status=cam_readblocklimits (path_id,target_id,target_lun,&minblksize,&maxblksize,&cam_status,&scsi_status);

	if ((status!=B_NO_ERROR)||(cam_status!=CAM_REQ_CMP)||(scsi_status!=0))
		return B_ERROR;

	ddprint ("successful\n");
	
	info->private_data=(void *)malloc(sizeof(struct driverdata));
	
	((struct driverdata *)info->private_data)->minblksize=minblksize;
	((struct driverdata *)info->private_data)->maxblksize=maxblksize;
	((struct driverdata *)info->private_data)->path_id=path_id;
	((struct driverdata *)info->private_data)->target_id=target_id;
	((struct driverdata *)info->private_data)->target_lun=target_lun;
	
	return B_NO_ERROR;
}

long ct_close (device_info *info)
{
	ddprint ("close\n");
	
	if (info->private_data!=NULL)
		free(info->private_data);
		
    return B_NO_ERROR;
}

long ct_control (device_info *info, ulong op, void *data)
{
	u_char path_id=((struct driverdata *)info->private_data)->path_id;
	u_char target_id=((struct driverdata *)info->private_data)->target_id;
	u_char target_lun=((struct driverdata *)info->private_data)->target_lun;
	
	u_char cam_status,scsi_status;
	
	long status;
	
	ddprint ("control (");
	ddprintlong (op);
	ddprint (")\n");	
	
	switch (op)
	{
		case TAPE_GETMINMAXBLKSIZE:
			((ulong *)data)[0]=((struct driverdata *)info->private_data)->minblksize;
			((ulong *)data)[1]=((struct driverdata *)info->private_data)->maxblksize;
			return status=B_NO_ERROR;
			break;

		case TAPE_REWIND:
		{
			ddprint ("rewinding...");

			status=cam_rewind (path_id,target_id,target_lun,&cam_status,&scsi_status);

			if ((status!=B_NO_ERROR)||(cam_status!=CAM_REQ_CMP)||(scsi_status!=0))
				status=B_ERROR;
			else
			{
				ddprint ("successful\n");
				status=B_NO_ERROR;
			}
			break;
		}
				
		case TAPE_ERASE:
		{
			ddprint ("rewinding...");

			status=cam_erase (path_id,target_id,target_lun,((uchar *)data)[0],&cam_status,&scsi_status);

			if ((status!=B_NO_ERROR)||(cam_status!=CAM_REQ_CMP)||(scsi_status!=0))
				status=B_ERROR;
			else
			{
				ddprint ("successful\n");
				status=B_NO_ERROR;
			}
			break;
		}

		default:
			status=B_ERROR;
			break;
	}
	
    return status;
}

long ct_read (device_info *info, void *buffer, ulong numBytes, ulong position)
{	
	u_char path_id=((struct driverdata *)info->private_data)->path_id;
	u_char target_id=((struct driverdata *)info->private_data)->target_id;
	u_char target_lun=((struct driverdata *)info->private_data)->target_lun;
	
	u_char cam_status,scsi_status;
	long status;

	status=cam_read (path_id,target_id,target_lun,buffer,numBytes,&cam_status,&scsi_status);

	if ((status!=B_NO_ERROR)||(cam_status!=CAM_REQ_CMP)||(scsi_status!=0))
		return B_ERROR;

    return numBytes;
}

long ct_write (device_info *info, void *buffer, ulong numBytes, ulong position)
{
	u_char path_id=((struct driverdata *)info->private_data)->path_id;
	u_char target_id=((struct driverdata *)info->private_data)->target_id;
	u_char target_lun=((struct driverdata *)info->private_data)->target_lun;
	
	u_char cam_status,scsi_status;
	long status;

	status=cam_write (path_id,target_id,target_lun,buffer,numBytes,&cam_status,&scsi_status);

	if ((status!=B_NO_ERROR)||(cam_status!=CAM_REQ_CMP)||(scsi_status!=0))
		return B_ERROR;
						
    return numBytes; 
}

#ifdef DEBUG

void ddprint (const char *string)
{
	port_id port;

	port=find_port ("DiagnosisPort");
	
	if (port==B_BAD_PORT_ID)
		return;
		
	(void)write_port (port,0,string,strlen(string));	
}

void ddprintlong (ulong x)
{
	char buf[20];
	sprintf (buf,"%ld",x);
	ddprint (buf);
}

void ddprinthexlong (ulong x)
{
	char buf[20];
	sprintf (buf,"0x%x",x);
	ddprint (buf);
}

void ddclear ()
{
	port_id port;

	port=find_port ("DiagnosisPort");
	
	if (port==B_BAD_PORT_ID)
		return;
		
	(void)write_port (port,1,NULL,0);	
}

#else

void ddprint (const char *string)
{
}

void ddprintlong (ulong x)
{
}

void ddprinthexlong (ulong x)
{
}

void ddclear ()
{
}

#endif

void outputautosensebuffer (struct sense_data *sensebuffer)
{
	if (sensebuffer->error_code & 128)
	{
		ddprint ("sensebuffer valid\n");
		
		if (((sensebuffer->error_code & 127)==0x70)||((sensebuffer->error_code & 127)==0x71))
		{
			if ((sensebuffer->error_code & 127)==0x70)
				ddprint ("current error\n");
			else if ((sensebuffer->error_code & 127)==0x71)
				ddprint ("deferred error\n");

			if (sensebuffer->filemarkEomIliReservedSensekey & 128)
				ddprint ("read file/setmark\n");
		
			if (sensebuffer->filemarkEomIliReservedSensekey & 64)
				ddprint ("end of medium\n");

			if (sensebuffer->filemarkEomIliReservedSensekey & 32)
				ddprint ("incorrect length\n");

			ddprint ("sense key :");
			switch (sensebuffer->filemarkEomIliReservedSensekey & 15)
			{
				case 0:
					ddprint ("NO SENSE");
					break;
				case 1:
					ddprint ("RECOVERED ERROR");
					break;
				case 2:
					ddprint ("NOT READY");
					break;
				case 3:
					ddprint ("MEDIUM ERROR");
					break;
				case 4:
					ddprint ("HARDWARE ERROR");
					break;
				case 5:
					ddprint ("ILLEGAL REQUEST");
					break;
				case 6:
					ddprint ("UNIT ATTENTION");
					break;
				case 7:
					ddprint ("DATA PROTECT");
					break;
				case 8:
					ddprint ("BLANK CHECK");
					break;
				case 9:
					ddprint ("<VendorSpecific>");
					break;
				case 0xa:
					ddprint ("COPY ABORTED");
					break;
				case 0xb:
					ddprint ("ABORTED COMMAND");
					break;
				case 0xc:
					ddprint ("EQUAL");
					break;
				case 0xd:
					ddprint ("VOLUME OVERFLOW");
					break;
				case 0xe:
					ddprint ("MISCOMPARE");
					break;
				case 0xf:
					ddprint ("RESERVED");
					break;					
				default:
					ddprint ("UNKNOWN");
			}
					
			ddprint ("\n");
		}
		else
			ddprint ("unknown error code\n");			
	}
}
